@@ -17,7 +17,7 @@ module Network.TLS
 
 	-- * Context object
 	, TLSCtx
-	, ctxHandle
+	, ctxConnection
 
 	-- * Creating a context
 	, client
@@ -34,10 +34,12 @@ module Network.TLS
 	-- * Crypto Key
 	, PrivateKey(..)
 	-- * Compressions & Predefined compressions
-	, Compression
+	, CompressionC(..)
+	, Compression(..)
 	, nullCompression
 	-- * Ciphers & Predefined ciphers
-	, Cipher
+	, Cipher(..)
+	, Bulk(..)
 	-- * Versions
 	, Version(..)
 	-- * Errors
@@ -46,6 +48,6 @@ module Network.TLS
 
 import Network.TLS.Struct (Version(..), TLSError(..))
 import Network.TLS.Crypto (PrivateKey(..))
-import Network.TLS.Cipher (Cipher(..))
-import Network.TLS.Compression (Compression(..), nullCompression)
+import Network.TLS.Cipher (Cipher(..), Bulk(..))
+import Network.TLS.Compression (CompressionC(..), Compression(..), nullCompression)
 import Network.TLS.Core
