@@ -63,6 +63,9 @@ import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as BC
 import qualified Data.ByteString.Lazy as L
 
+import qualified Crypto.Hash.SHA1 as SHA1
+import qualified Crypto.Hash.MD5 as MD5
+
 data CurrentParams = CurrentParams
 	{ cParamsVersion     :: Version               -- ^ current protocol version
 	, cParamsKeyXchgType :: CipherKeyExchangeType -- ^ current key exchange type
@@ -105,11 +108,7 @@ getHandshakeType = do
  - decode and encode headers
  -}
 decodeHeader :: ByteString -> Either TLSError Header
-decodeHeader = runGetErr "header" $ do
-	ty  <- getHeaderType
-	v   <- getVersion
-	len <- getWord16
-	return $ Header ty v len
+decodeHeader = runGetErr "header" $ liftM3 Header getHeaderType getVersion getWord16
 
 encodeHeader :: Header -> ByteString
 encodeHeader (Header pt ver len) = runPut (putHeaderType pt >> putVersion ver >> putWord16 len)
@@ -141,7 +140,7 @@ decodeAlerts = runGetErr "alerts" $ loop
 
 encodeAlerts :: [(AlertLevel, AlertDescription)] -> ByteString
 encodeAlerts l = runPut $ mapM_ encodeAlert l
-	where encodeAlert (al, ad) = (putWord8 (valOfType al) >> putWord8 (valOfType ad))
+	where encodeAlert (al, ad) = putWord8 (valOfType al) >> putWord8 (valOfType ad)
 
 {- decode and encode HANDSHAKE -}
 decodeHandshakeHeader :: Get (HandshakeType, Bytes)
@@ -207,8 +206,7 @@ decodeServerHelloDone = return ServerHelloDone
 
 decodeCertificates :: Get Handshake
 decodeCertificates = do
-	certslen <- getWord24
-	certs <- getCerts certslen >>= return . map (decodeCertificate . L.fromChunks . (:[]))
+	certs <- getWord24 >>= getCerts >>= return . map (decodeCertificate . L.fromChunks . (:[]))
 	let (l, r) = partitionEithers certs
 	if length l > 0
 		then fail ("error certificate parsing: " ++ show l)
@@ -219,13 +217,15 @@ decodeFinished = do
 	opaque <- remaining >>= getBytes
 	return $ Finished $ opaque
 
-getSignatureHashAlgorithm :: Int -> Get [ (HashAlgorithm, SignatureAlgorithm) ]
-getSignatureHashAlgorithm 0   = return []
-getSignatureHashAlgorithm len = do
+getSignatureHashAlgorithm :: Get (HashAlgorithm, SignatureAlgorithm)
+getSignatureHashAlgorithm = do
 	h <- fromJust . valToType <$> getWord8
 	s <- fromJust . valToType <$> getWord8
-	xs <- getSignatureHashAlgorithm (len - 2)
-	return ((h, s) : xs)
+	return (h,s)
+
+getSignatureHashAlgorithms :: Int -> Get [ (HashAlgorithm, SignatureAlgorithm) ]
+getSignatureHashAlgorithms 0   = return []
+getSignatureHashAlgorithms len = liftM2 (:) getSignatureHashAlgorithm (getSignatureHashAlgorithms (len-2))
 
 decodeCertRequest :: CurrentParams -> Get Handshake
 decodeCertRequest cp = do
@@ -234,7 +234,7 @@ decodeCertRequest cp = do
 	sigHashAlgs <- if cParamsVersion cp >= TLS12
 		then do
 			sighashlen <- getWord16
-			Just <$> getSignatureHashAlgorithm (fromIntegral sighashlen)
+			Just <$> getSignatureHashAlgorithms (fromIntegral sighashlen)
 		else return Nothing
 	dNameLen <- getWord16
 	when (cParamsVersion cp < TLS12 && dNameLen < 3) $ fail "certrequest distinguishname not of the correct size"
@@ -372,10 +372,8 @@ getSession = do
 		len -> Session . Just <$> getBytes len
 
 putSession :: Session -> Put
-putSession (Session session) =
-	case session of
-		Nothing -> putWord8 0
-		Just s  -> putWord8 (fromIntegral $ B.length s) >> putBytes s
+putSession (Session Nothing)  = putWord8 0
+putSession (Session (Just s)) = putWord8 (fromIntegral $ B.length s) >> putBytes s
 
 getCerts :: Int -> Get [Bytes]
 getCerts 0   = return []
@@ -446,61 +444,72 @@ encodeExtSecureRenegotiation cvd msvd = runPut $ do
 {-
  - generate things for packet content
  -}
-generateMasterSecret_TLS, generateMasterSecret_SSL :: Bytes -> ClientRandom -> ServerRandom -> Bytes
-generateMasterSecret_TLS premasterSecret (ClientRandom c) (ServerRandom s) =
-	prf_MD5SHA1 premasterSecret seed 48
-	where
-		seed = B.concat [ "master secret", c, s ]
+type PRF = Bytes -> Bytes -> Int -> Bytes
 
+generateMasterSecret_SSL :: Bytes -> ClientRandom -> ServerRandom -> Bytes
 generateMasterSecret_SSL premasterSecret (ClientRandom c) (ServerRandom s) =
 	B.concat $ map (computeMD5) ["A","BB","CCC"]
 	where
-		computeMD5  label = hashMD5 $ B.concat [ premasterSecret, computeSHA1 label ]
-		computeSHA1 label = hashSHA1 $ B.concat [ label, premasterSecret, c, s ]
+		computeMD5  label = MD5.hash $ B.concat [ premasterSecret, computeSHA1 label ]
+		computeSHA1 label = SHA1.hash $ B.concat [ label, premasterSecret, c, s ]
+
+generateMasterSecret_TLS :: PRF -> Bytes -> ClientRandom -> ServerRandom -> Bytes
+generateMasterSecret_TLS prf premasterSecret (ClientRandom c) (ServerRandom s) =
+	prf premasterSecret seed 48
+	where
+		seed = B.concat [ "master secret", c, s ]
 
 generateMasterSecret :: Version -> Bytes -> ClientRandom -> ServerRandom -> Bytes
-generateMasterSecret ver =
-	if ver < TLS10 then generateMasterSecret_SSL else generateMasterSecret_TLS
+generateMasterSecret SSL2  = generateMasterSecret_SSL
+generateMasterSecret SSL3  = generateMasterSecret_SSL
+generateMasterSecret TLS10 = generateMasterSecret_TLS prf_MD5SHA1
+generateMasterSecret TLS11 = generateMasterSecret_TLS prf_MD5SHA1
+generateMasterSecret TLS12 = generateMasterSecret_TLS prf_SHA256
 
-generateKeyBlock_TLS :: ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
-generateKeyBlock_TLS (ClientRandom c) (ServerRandom s) mastersecret kbsize =
-	prf_MD5SHA1 mastersecret seed kbsize
-	where
-		seed = B.concat [ "key expansion", s, c ]
+generateKeyBlock_TLS :: PRF -> ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
+generateKeyBlock_TLS prf (ClientRandom c) (ServerRandom s) mastersecret kbsize =
+	prf mastersecret seed kbsize where seed = B.concat [ "key expansion", s, c ]
 
 generateKeyBlock_SSL :: ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
 generateKeyBlock_SSL (ClientRandom c) (ServerRandom s) mastersecret kbsize =
 	B.concat $ map computeMD5 $ take ((kbsize `div` 16) + 1) labels
 	where
 		labels            = [ uncurry BC.replicate x | x <- zip [1..] ['A'..'Z'] ]
-		computeMD5  label = hashMD5 $ B.concat [ mastersecret, computeSHA1 label ]
-		computeSHA1 label = hashSHA1 $ B.concat [ label, mastersecret, s, c ]
+		computeMD5  label = MD5.hash $ B.concat [ mastersecret, computeSHA1 label ]
+		computeSHA1 label = SHA1.hash $ B.concat [ label, mastersecret, s, c ]
 
 generateKeyBlock :: Version -> ClientRandom -> ServerRandom -> Bytes -> Int -> Bytes
-generateKeyBlock ver =
-	if ver < TLS10 then generateKeyBlock_SSL else generateKeyBlock_TLS
+generateKeyBlock SSL2  = generateKeyBlock_SSL
+generateKeyBlock SSL3  = generateKeyBlock_SSL
+generateKeyBlock TLS10 = generateKeyBlock_TLS prf_MD5SHA1
+generateKeyBlock TLS11 = generateKeyBlock_TLS prf_MD5SHA1
+generateKeyBlock TLS12 = generateKeyBlock_TLS prf_SHA256
 
-generateFinished_TLS :: Bytes -> Bytes -> HashCtx -> HashCtx -> Bytes
-generateFinished_TLS label mastersecret md5ctx sha1ctx =
-	prf_MD5SHA1 mastersecret seed 12
+generateFinished_TLS :: PRF -> Bytes -> Bytes -> HashCtx -> Bytes
+generateFinished_TLS prf label mastersecret hashctx = prf mastersecret seed 12
 	where
-		seed = B.concat [ label, finalizeHash md5ctx, finalizeHash sha1ctx ]
+		seed = B.concat [ label, hashFinal hashctx ]
 
-generateFinished_SSL :: Bytes -> Bytes -> HashCtx -> HashCtx -> Bytes
-generateFinished_SSL sender mastersecret md5ctx sha1ctx =
-	B.concat [md5hash, sha1hash]
+generateFinished_SSL :: Bytes -> Bytes -> HashCtx -> Bytes
+generateFinished_SSL sender mastersecret hashctx = B.concat [md5hash, sha1hash]
 	where
-		md5hash  = hashMD5 $ B.concat [ mastersecret, pad2, md5left ]
-		sha1hash = hashSHA1 $ B.concat [ mastersecret, B.take 40 pad2, sha1left ]
-		md5left  = finalizeHash $ foldl updateHash md5ctx [ sender, mastersecret, pad1 ]
-		sha1left = finalizeHash $ foldl updateHash sha1ctx [ sender, mastersecret, B.take 40 pad1 ]
+		md5hash  = MD5.hash $ B.concat [ mastersecret, pad2, md5left ]
+		sha1hash = SHA1.hash $ B.concat [ mastersecret, B.take 40 pad2, sha1left ]
+
+		lefthash = hashFinal $ flip hashUpdateSSL (pad1, B.take 40 pad1)
+		                     $ foldl hashUpdate hashctx [sender,mastersecret]
+		(md5left,sha1left) = B.splitAt 16 lefthash
 		pad2     = B.replicate 48 0x5c
 		pad1     = B.replicate 48 0x36
 
-generateClientFinished :: Version -> Bytes -> HashCtx -> HashCtx -> Bytes
-generateClientFinished ver =
-	if ver < TLS10 then generateFinished_SSL "CLNT" else generateFinished_TLS "client finished"
+generateClientFinished :: Version -> Bytes -> HashCtx -> Bytes
+generateClientFinished ver
+	| ver < TLS10 = generateFinished_SSL "CLNT"
+	| ver < TLS12 = generateFinished_TLS prf_MD5SHA1 "client finished"
+	| otherwise   = generateFinished_TLS prf_SHA256 "client finished"
 
-generateServerFinished :: Version -> Bytes -> HashCtx -> HashCtx -> Bytes
-generateServerFinished ver =
-	if ver < TLS10 then generateFinished_SSL "SRVR" else generateFinished_TLS "server finished"
+generateServerFinished :: Version -> Bytes -> HashCtx -> Bytes
+generateServerFinished ver
+	| ver < TLS10 = generateFinished_SSL "SRVR"
+	| ver < TLS12 = generateFinished_TLS prf_MD5SHA1 "server finished"
+	| otherwise   = generateFinished_TLS prf_SHA256 "server finished"
