@@ -16,7 +16,6 @@ module Network.TLS.Struct
 	, CipherType(..)
 	, CipherData(..)
 	, Extension
-	, EncryptedData(..)
 	, CertificateType(..)
 	, HashAlgorithm(..)
 	, SignatureAlgorithm(..)
@@ -48,7 +47,7 @@ module Network.TLS.Struct
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as B (length)
 import Data.Word
-import Data.Certificate.X509
+import Data.Certificate.X509 (X509)
 import Data.Typeable
 import Control.Monad.Error (Error(..))
 import Control.Exception (Exception(..))
@@ -137,7 +136,7 @@ data Packet =
 	| AppData ByteString
 	deriving (Show,Eq)
 
-data Header = Header ProtocolType Version Word16 deriving (Show, Eq)
+data Header = Header ProtocolType Version Word16 deriving (Show,Eq)
 
 newtype ServerRandom = ServerRandom Bytes deriving (Show, Eq)
 newtype ClientRandom = ClientRandom Bytes deriving (Show, Eq)
@@ -157,9 +156,6 @@ serverRandom l = constrRandom32 ServerRandom l
 clientRandom :: Bytes -> Maybe ClientRandom
 clientRandom l = constrRandom32 ClientRandom l
 
-newtype EncryptedData = EncryptedData ByteString
-	deriving (Show)
-
 data AlertLevel =
 	  AlertLevel_Warning
 	| AlertLevel_Fatal
